# Generated by Django 3.2.19 on 2023-10-24 12:13

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0010_identity'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='identity',
            name='profile',
        ),
        migrations.AddField(
            model_name='identity',
            name='claims',
            field=models.JSONField(default=dict),
        ),
    ]
