<div id="file-contents" class="card">
    <div class="card-header">
        {{ file_in_artifact.path }}
        {% if not file_in_artifact.complete %}(incomplete){% endif %}
    </div>
    {% if file_in_artifact.complete %}
        <div class="card-body text-white bg-dark">
            {% block file_contents %}{% endblock %}
        </div>
    {% endif %}
</div>
