# Generated by Django 3.2.19 on 2024-07-29 17:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0068_collection_workflow_onetoone'),
    ]

    operations = [
        migrations.AddField(
            model_name='collectionitem',
            name='parent_category',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
    ]
