# Generated by Django 3.2.19 on 2024-09-05 08:42

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0077_wait_task_type'),
    ]

    operations = [
        migrations.AddField(
            model_name='collectionitem',
            name='created_by_workflow',
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name='workflow_created_%(class)s',
                to='db.workrequest',
            ),
        ),
        migrations.AddField(
            model_name='collectionitem',
            name='removed_by_workflow',
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name='workflow_removed_%(class)s',
                to='db.workrequest',
            ),
        ),
    ]
