# Generated by Django 3.2.19 on 2024-02-26 13:47

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0017_collections'),
    ]

    operations = [
        migrations.AlterField(
            model_name='workrequest',
            name='worker',
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name='assigned_work_requests',
                to='db.worker',
            ),
        ),
    ]
