# Generated by Django 3.2.19 on 2024-10-10 20:45

from django.db import migrations, models

import debusine.db.models.auth


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0088_roles_for_scopes'),
    ]

    operations = [
        migrations.AlterField(
            model_name='group',
            name='name',
            field=models.CharField(
                max_length=255,
                validators=[debusine.db.models.auth.validate_group_name],
            ),
        ),
    ]
