# Generated by Django 3.2.19 on 2024-09-19 08:40

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0083_set_fallback_scope'),
    ]

    operations = [
        migrations.AlterField(
            model_name='workspace',
            name='name',
            field=models.CharField(max_length=255),
        ),
        migrations.AlterField(
            model_name='workspace',
            name='scope',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name='workspaces',
                to='db.scope',
            ),
        ),
        migrations.AddConstraint(
            model_name='workspace',
            constraint=models.UniqueConstraint(
                fields=('scope', 'name'), name='db_workspace_unique_scope_name'
            ),
        ),
    ]
