# Generated by Django 3.2.12 on 2023-05-31 07:17

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0005_workrequest_created_by'),
    ]

    operations = [
        migrations.AlterField(
            model_name='workrequest',
            name='created_by',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]
