<button class="btn btn-primary btn-sm"
        type="button"
        data-bs-toggle="collapse"
        data-bs-target="#{{ key }}Div">Show / Hide Token</button>
<div class="collapse" id="{{ key }}Div">
    <code>{{ key }}</code>
</div>
