{% extends "web/_base.html" %}
{% block title %}List of work requests{% endblock %}
{% block content %}
    <h1>List of work requests</h1>
    <p>
        <a href="https://freexian-team.pages.debian.net/debusine/explanation/concepts.html#work-requests">Work Requests</a>
        are how
        <a href="https://freexian-team.pages.debian.net/debusine/explanation/concepts.html#tasks">tasks</a>
        are scheduled to be processed by workers. Here you can see a list of
        current and past work requests processed by Debusine.
    </p>
    {% include "web/_work_request-list.html" %}
    {% include "web/_pagination.html" %}
{% endblock %}
