{{ worker.name }}
{% if worker.worker_type == "external" %}
    (<abbr title="External">E</abbr>)
{% elif worker.worker_type == "celery" %}
    (<abbr title="Celery">C</abbr>)
{% elif worker.worker_type == "signing" %}
    (<abbr title="Signing">S</abbr>)
{% else %}
    ({{ worker.worker_type }})
{% endif %}
