# Generated by Django 4.2.11 on 2024-04-04 15:36

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import debusine.tasks.models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0029_workflowtemplate'),
    ]

    operations = [
        migrations.AlterField(
            model_name='workrequest',
            name='task_type',
            field=models.CharField(
                choices=[
                    (
                        debusine.tasks.models.TaskTypes['WORKER'],
                        debusine.tasks.models.TaskTypes['WORKER'],
                    ),
                    (
                        debusine.tasks.models.TaskTypes['SERVER'],
                        debusine.tasks.models.TaskTypes['SERVER'],
                    ),
                    (
                        debusine.tasks.models.TaskTypes['INTERNAL'],
                        debusine.tasks.models.TaskTypes['INTERNAL'],
                    ),
                    (
                        debusine.tasks.models.TaskTypes['WORKFLOW'],
                        debusine.tasks.models.TaskTypes['WORKFLOW'],
                    ),
                ],
                default=debusine.tasks.models.TaskTypes['WORKER'],
                editable=False,
                max_length=8,
                verbose_name='Type of task to execute',
            ),
        ),
    ]
