{% extends "web/_base.html" %}
{% load debusine %}
{% block content %}
    <h1>{{ title }}</h1>
    <form method="post">
        {% csrf_token %}
        {{ form.as_p }}
        <input class="btn btn-primary btn-sm" type="submit" value="Save">
    </form>
{% endblock %}
