# Generated by Django 4.2.15 on 2024-10-25 15:41

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0092_default_workspace_public'),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name='scoperole',
            name='db_scoperole_unique_scope_group_role',
        ),
        migrations.RenameField(
            model_name='scoperole',
            old_name='scope',
            new_name='resource',
        ),
        migrations.AddConstraint(
            model_name='scoperole',
            constraint=models.UniqueConstraint(
                fields=('resource', 'group', 'role'),
                name='db_scoperole_unique_resource_group_role',
            ),
        ),
    ]
