{% extends "web/_base.html" %}
{% block title %}List of workers{% endblock %}
{% block content %}
    <h1>List of workers</h1>
    <p>This is a list of all the registered workers and their running work requests.</p>
    {% include "web/_worker-list.html" %}
    {% include "web/_pagination.html" %}
{% endblock %}
