# Generated by Django 3.2.19 on 2024-10-04 16:57

import django.core.serializers.json
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0086_add_scoped_group'),
    ]

    operations = [
        migrations.AlterField(
            model_name='workflowtemplate',
            name='task_data',
            field=models.JSONField(
                blank=True,
                default=dict,
                encoder=django.core.serializers.json.DjangoJSONEncoder,
            ),
        ),
        migrations.AlterField(
            model_name='workrequest',
            name='dynamic_task_data',
            field=models.JSONField(
                blank=True,
                encoder=django.core.serializers.json.DjangoJSONEncoder,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name='workrequest',
            name='event_reactions_json',
            field=models.JSONField(
                blank=True,
                db_column='event_reactions',
                default=dict,
                encoder=django.core.serializers.json.DjangoJSONEncoder,
            ),
        ),
        migrations.AlterField(
            model_name='workrequest',
            name='task_data',
            field=models.JSONField(
                blank=True,
                default=dict,
                encoder=django.core.serializers.json.DjangoJSONEncoder,
            ),
        ),
        migrations.AlterField(
            model_name='workrequest',
            name='workflow_data_json',
            field=models.JSONField(
                blank=True,
                db_column='workflow_data',
                default=dict,
                encoder=django.core.serializers.json.DjangoJSONEncoder,
            ),
        ),
    ]
