{% extends "web/_base.html" %}
{% load debusine %}
{% block title %}{{ title }}{% endblock %}
{% block content %}
    <h1>Workflow template {{ workflow_template.name }}</h1>
    <div class="card mb-2" id="workflow-information">
        <div class="card-header">
            <b>{{ workflow_template.name }}</b> (type <b>{{ workflow_template.task_name }}</b>, priority {{ workflow_template.priority }})
        </div>
        <div class="card-body text-white bg-dark">{{ workflow_template.task_data|format_yaml }}</div>
    </div>
{% endblock %}
