{% extends "web/_base.html" %}
{% block title %}New token created{% endblock %}
{% block content %}
    <h1>New token created</h1>
    <ul>
        <li>
            Token key:
            {% include "web/_token-collapse.html" with key=token.key only %}
        </li>
        <li>Comment: {{ token.comment }}</li>
        <li>
            Enabled:
            {% include "web/_check_icon.html" with value=token.enabled only %}
        </li>
        <li>Created at: {{ token.created_at }}</li>
    </ul>
    <p>Save the token key. If you lose it, you can generate a new one.</p>
    <a class="btn btn-primary btn-sm"
       href="{% url "user:token-list" username=person.username %}">Token list</a>
{% endblock %}
