# Generated by Django 4.2.18 on 2025-02-05 11:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        (
            'db',
            '0124_drop_category_from_db_asset_unique_signing_key_fingerprints',
        ),
    ]

    operations = [
        migrations.AddField(
            model_name='group',
            name='ephemeral',
            field=models.BooleanField(
                default=False,
                help_text='remove the group if it has no roles assigned',
            ),
        ),
    ]
