# Generated by Django 4.2.11 on 2024-05-01 05:16

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0043_create_system_user'),
    ]

    operations = [
        migrations.AddField(
            model_name='collection',
            name='retains_artifacts',
            field=models.BooleanField(default=False),
        ),
    ]
