# Generated by Django 4.2.15 on 2024-12-09 22:00

from django.db import migrations, models

from debusine.db.migrations._utils import ReloadModel


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0103_cascade_deletion_on_role_assignments'),
    ]

    operations = [
        migrations.AlterField(
            model_name='workspacerole',
            name='role',
            field=models.CharField(
                choices=[('owner', 'Owner'), ('contributor', 'Contributor')],
                max_length=16,
            ),
        ),
        ReloadModel(("db", "collectionitemmatchconstraint")),
    ]
