# Generated by Django 3.2.19 on 2024-05-28 16:14

from django.db import migrations

from debusine.db.migrations._utils import make_work_request_task_renamer
from debusine.tasks.models import TaskTypes


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0061_workrequest_supersedes'),
    ]

    operations = [
        make_work_request_task_renamer(
            TaskTypes.SERVER, "internalnoop", "servernoop"
        ),
    ]
