# Generated by Django 4.2.18 on 2025-02-06 15:38

import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        (
            'db',
            '0126_filestore_instance_wide',
        ),
    ]

    operations = [
        migrations.AddField(
            model_name='workspace',
            name='created_at',
            field=models.DateTimeField(
                auto_now_add=True,
                default=django.utils.timezone.now,
                help_text='time the workspace was created',
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='workspace',
            name='expiration_delay',
            field=models.DurationField(
                blank=True,
                help_text='if set, time since the last task completion timeafter which the workspace can be deleted',
                null=True,
            ),
        ),
    ]
