{% extends "web/_base.html" %}
{% block title %}{{ action }}Create an artifact{% endblock %}
{% block content %}
    <h1>Create artifact</h1>
    <p>
        Artifacts are at the heart of Debusine. Artifacts are both inputs
        (submitted by users) and outputs (generated by tasks). See the
        <a href="https://freexian-team.pages.debian.net/debusine/explanation/concepts.html#artifacts">artifacts introduction</a>
        and the
        <a href="https://freexian-team.pages.debian.net/debusine/reference/artifacts.html">full artifacts reference</a>
        to learn more.
    </p>
    <p>
        Artifacts are usually created by
        <a href="https://freexian-team.pages.debian.net/debusine/reference/debusine-cli.html">debusine-client</a>
        or generated as results of
        <a href="https://freexian-team.pages.debian.net/debusine/reference/tasks.html">tasks</a>
        executed by Debusine. This page allows you to create a artifact manually as
        a set of files and YAML metadata, and is intended to be used mostly as a
        low-level debugging/testing tool.
    </p>
    <form method="post" enctype="multipart/form-data">
        {% csrf_token %}
        {{ form.as_p }}
        <input class="btn btn-primary btn-sm" type="submit" value="Create">
    </form>
    <br>
{% endblock %}
