# Generated by Django 3.2.19 on 2024-04-12 10:48

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0031_workrequest_priorities'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='workrequest',
            options={
                'permissions': [
                    (
                        'manage_workrequest_priorities',
                        'Can set positive priority adjustments on work requests',
                    )
                ]
            },
        ),
    ]
